/*
 * $Id: glist.h,v 1.1 1998/01/05 09:20:12 dhall Exp $
 *
 * glist.h - special generic list macros for double-linked lists
 *
 * This file is part of libswiss
 *
 * Copyright (c) 1997 by Vi'Rage Studios
 *
 */

#if !defined(_GLIST_H)
#define _GLIST_H

typedef struct generic_list_data	list_data;

struct generic_list_data {
   size_t size;

   void *head;
   void *tail;
};

/* NOTE: GCC seems to fatal on ()'s around struct pointer members */

/* 'insert' node via the list->head pointer */
#define GL_INSERT(node, type, list, next, prev) \
do {						\
   if ((list)->size == 0) {			\
      (list)->head = (list)->tail = (node);	\
      (node)->next = (node)->prev = NULL;	\
   } else {					\
      (node)->next = (list)->head;		\
      (node)->prev = NULL;			\
      ((type *) (list)->head)->prev = (node);	\
      (list)->head = (node);			\
   }						\
   (list)->size++;				\
} while (0)

/* 'append' node via the list->tail pointer */
#define GL_APPEND(node, type, list, next, prev)	\
do {						\
   if ((list)->size == 0) {			\
      (list)->head = (list)->tail = (node);	\
      (node)->next = (node)->prev = NULL;	\
   } else {					\
      (node)->prev = (list)->tail;		\
      (node)->next = NULL;			\
      ((type *) (list)->tail)->next = (node);	\
      (list)->tail = (node);			\
   }						\
   (list)->size++;				\
} while (0)

/* 'delete' node from list */
#define GL_DELETE(node, list, next, prev)	\
do {						\
   if ((node)->next)				\
      (node)->next->prev = (node)->prev;	\
   if ((node)->prev)				\
      (node)->prev->next = (node)->next;	\
   if ((list)->head == (node))			\
      (list)->head = (node)->next;		\
   if ((list)->tail == (node))			\
      (list)->tail = (node)->prev;		\
   if ((list)->size > 0)			\
      (list)->size--;				\
   else						\
      glist_error ((node));			\
} while (0)

list_data *glist_create (void);
void glist_error (void *);

#endif /* _GLIST_H */
